from __future__ import annotations

import inspect
from collections.abc import AsyncGenerator, Callable, Generator
from types import TracebackType
from typing import Any, cast

import pytest

from diwire import Container, Lifetime, LockMode, Scope
from diwire._internal.resolvers.assembly.compiler import ResolversAssemblyCompiler
from diwire.exceptions import DIWireAsyncDependencyInSyncContextError, DIWireScopeMismatchError


class _MatrixService:
    def __init__(self) -> None:
        self.value: object | None = None


class _MatrixResource:
    pass


class _MatrixAsyncDependency:
    pass


class _SignatureService:
    def __init__(self, payload: object) -> None:
        self.payload = payload


def _build_resolver_with_cleanup_mode(
    *,
    container: Container,
    cleanup_enabled: bool,
) -> Any:
    return ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
        cleanup_enabled=cleanup_enabled,
    )


@pytest.mark.parametrize(
    ("provider_kind", "lifetime", "scope", "expect_same", "expect_call_count"),
    [
        ("instance", None, Scope.APP, True, None),
        ("concrete", Lifetime.TRANSIENT, Scope.APP, False, None),
        ("factory", Lifetime.SCOPED, Scope.APP, True, 1),
        ("factory", Lifetime.TRANSIENT, Scope.REQUEST, False, 2),
        ("factory", Lifetime.SCOPED, Scope.REQUEST, True, 1),
    ],
)
def test_assembly_matrix_caching_identity_by_kind_lifetime_scope(
    provider_kind: str,
    lifetime: Lifetime | None,
    scope: Scope,
    expect_same: Any,
    expect_call_count: int | None,
) -> None:
    calls = 0

    def build_service() -> _MatrixService:
        nonlocal calls
        calls += 1
        service = _MatrixService()
        service.value = calls
        return service

    container = Container()
    if provider_kind == "instance":
        instance = _MatrixService()
        instance.value = "instance"
        container.add_instance(instance, provides=_MatrixService)
    elif provider_kind == "concrete":
        assert lifetime is not None
        container.add(
            _MatrixService,
            provides=_MatrixService,
            lifetime=lifetime,
            scope=scope,
        )
    else:
        assert lifetime is not None
        container.add_factory(
            build_service,
            provides=_MatrixService,
            lifetime=lifetime,
            scope=scope,
        )

    if scope is Scope.APP:
        first = container.resolve(_MatrixService)
        second = container.resolve(_MatrixService)
    else:
        with container.enter_scope() as request_scope:
            first = request_scope.resolve(_MatrixService)
            second = request_scope.resolve(_MatrixService)

    assert (first is second) is bool(expect_same)
    if expect_call_count is not None:
        assert calls == expect_call_count


def test_assembly_matrix_scoped_cache_isolated_across_scope_instances() -> None:
    calls = 0

    def build_service() -> _MatrixService:
        nonlocal calls
        calls += 1
        service = _MatrixService()
        service.value = calls
        return service

    container = Container()
    container.add_factory(
        build_service,
        provides=_MatrixService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as first_scope:
        first = first_scope.resolve(_MatrixService)
        assert first is first_scope.resolve(_MatrixService)

    with container.enter_scope() as second_scope:
        second = second_scope.resolve(_MatrixService)
        assert second is second_scope.resolve(_MatrixService)

    assert first is not second
    assert calls == 2


@pytest.mark.parametrize("provider_kind", ["generator", "context_manager"])
@pytest.mark.parametrize("cleanup_enabled", [True, False])
def test_assembly_matrix_cleanup_behavior_respects_cleanup_enabled(
    provider_kind: str,
    cleanup_enabled: Any,
) -> None:
    events: list[str] = []
    exit_calls = 0

    def provide_generator() -> Generator[_MatrixResource, None, None]:
        events.append("enter")
        try:
            yield _MatrixResource()
        finally:
            events.append("exit")

    class _ManagedContext:
        def __enter__(self) -> _MatrixResource:
            events.append("enter")
            return _MatrixResource()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            nonlocal exit_calls
            exit_calls += 1
            events.append("exit")

    def provide_context_manager() -> _ManagedContext:
        return _ManagedContext()

    container = Container()
    if provider_kind == "generator":
        container.add_generator(
            provide_generator,
            provides=_MatrixResource,
            scope=Scope.REQUEST,
            lifetime=Lifetime.SCOPED,
        )
    else:
        container.add_context_manager(
            provide_context_manager,
            provides=_MatrixResource,
            scope=Scope.REQUEST,
            lifetime=Lifetime.SCOPED,
        )

    root_resolver = _build_resolver_with_cleanup_mode(
        container=container,
        cleanup_enabled=bool(cleanup_enabled),
    )
    request_scope = root_resolver.enter_scope()
    resolved = request_scope.resolve(_MatrixResource)

    assert isinstance(resolved, _MatrixResource)
    assert "enter" in events
    expected_callbacks = 1 if bool(cleanup_enabled) else 0
    assert len(request_scope._cleanup_callbacks) == expected_callbacks
    request_scope.__exit__(None, None, None)

    if provider_kind == "context_manager":
        assert exit_calls == expected_callbacks
    elif bool(cleanup_enabled):
        assert "exit" in events


def test_assembly_matrix_scope_mismatch_for_request_scoped_dependency_at_root() -> None:
    container = Container()
    container.add_factory(
        _MatrixService,
        provides=_MatrixService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(DIWireScopeMismatchError, match="requires opened scope level"):
        container.resolve(_MatrixService)


@pytest.mark.asyncio
async def test_assembly_matrix_async_dependency_chain_requires_aresolve() -> None:
    async def provide_dependency() -> AsyncGenerator[_MatrixAsyncDependency, None]:
        try:
            yield _MatrixAsyncDependency()
        finally:
            pass

    def build_consumer(dependency: _MatrixAsyncDependency) -> _MatrixService:
        service = _MatrixService()
        service.value = dependency
        return service

    container = Container()
    container.add_generator(
        provide_dependency,
        provides=_MatrixAsyncDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_factory(
        build_consumer,
        provides=_MatrixService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        with pytest.raises(
            DIWireAsyncDependencyInSyncContextError,
            match="requires asynchronous resolution",
        ):
            request_scope.resolve(_MatrixService)

    async with container.enter_scope() as request_scope:
        resolved = await request_scope.aresolve(_MatrixService)
        assert isinstance(resolved.value, _MatrixAsyncDependency)


@pytest.mark.parametrize(
    ("lock_mode", "has_thread_lock"),
    [
        (LockMode.THREAD, True),
        (LockMode.NONE, False),
    ],
)
def test_assembly_matrix_cached_sync_path_lock_generation_follows_lock_mode(
    lock_mode: LockMode,
    has_thread_lock: Any,
) -> None:
    calls = 0

    def build_service() -> _MatrixService:
        nonlocal calls
        calls += 1
        return _MatrixService()

    container = Container()
    container.add_factory(
        build_service,
        provides=_MatrixService,
        lifetime=Lifetime.SCOPED,
        lock_mode=lock_mode,
    )
    root_resolver = container.compile()

    first = root_resolver.resolve(_MatrixService)
    second = root_resolver.resolve(_MatrixService)
    assert first is second
    assert calls == 1
    assert bool(has_thread_lock) is (lock_mode is LockMode.THREAD)


@pytest.mark.parametrize("signature_kind", ["positional", "positional_only", "keyword_only"])
def test_assembly_matrix_signature_wiring_for_required_parameters(signature_kind: str) -> None:
    def positional(value: int) -> _SignatureService:
        return _SignatureService(value)

    def positional_only(value: int, /) -> _SignatureService:
        return _SignatureService(value)

    def keyword_only(*, value: int) -> _SignatureService:
        return _SignatureService(value)

    builders: dict[str, Callable[..., _SignatureService]] = {
        "positional": positional,
        "positional_only": positional_only,
        "keyword_only": keyword_only,
    }
    builder = builders[signature_kind]
    signature = inspect.signature(builder)

    container = Container()
    container.add_instance(42, provides=int)
    container.add_factory(
        builder,
        provides=_SignatureService,
        dependencies={
            int: signature.parameters["value"],
        },
    )

    resolved = container.resolve(_SignatureService)

    assert resolved.payload == 42


@pytest.mark.parametrize("signature_kind", ["var_positional", "var_keyword"])
def test_assembly_matrix_signature_wiring_for_variadic_parameters(signature_kind: str) -> None:
    def var_positional(*values: int) -> _SignatureService:
        return _SignatureService(tuple(values))

    def var_keyword(**options: int) -> _SignatureService:
        return _SignatureService(dict(options))

    builders: dict[str, Callable[..., _SignatureService]] = {
        "var_positional": var_positional,
        "var_keyword": var_keyword,
    }
    builder = builders[signature_kind]
    signature = inspect.signature(builder)

    container = Container()
    values_type: Any
    if signature_kind == "var_positional":
        values_type = tuple[int, ...]
        payload: object = (1, 2, 3)
        parameter = signature.parameters["values"]
    else:
        values_type = dict[str, int]
        payload = {"first": 1, "second": 2}
        parameter = signature.parameters["options"]

    container.add_instance(cast("Any", payload), provides=values_type)
    container.add_factory(
        builder,
        provides=_SignatureService,
        dependencies={
            values_type: parameter,
        },
    )

    resolved = container.resolve(_SignatureService)

    assert resolved.payload == payload


@pytest.mark.asyncio
async def test_assembly_matrix_sync_only_graph_has_sync_async_parity() -> None:
    container = Container()
    container.add_factory(
        _sync_only_service,
        provides=_MatrixService,
        lifetime=Lifetime.SCOPED,
    )

    sync_resolved = container.resolve(_MatrixService)
    async_resolved = await container.aresolve(_MatrixService)

    assert sync_resolved is async_resolved
    assert async_resolved.value == "sync-only"


def _sync_only_service() -> _MatrixService:
    service = _MatrixService()
    service.value = "sync-only"
    return service



from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path

import pytest

_ENABLED_ENV = "TYPER_E2E_ENABLED"
_REPO_ROOT = Path(__file__).resolve().parents[4]

if os.getenv(_ENABLED_ENV) != "1":
    pytest.skip(
        f"{_ENABLED_ENV} is not set to 1; skipping docker-compose Typer e2e tests.",
        allow_module_level=True,
    )


def _run_cli(*args: str) -> subprocess.CompletedProcess[str]:
    # Safe static command invocation for local test fixture module.
    return subprocess.run(  # noqa: S603
        [sys.executable, "-m", "tests.e2e.typer.setup.app", *args],
        cwd=_REPO_ROOT,
        check=False,
        capture_output=True,
        text=True,
        timeout=20,
    )


def test_context_direct_returns_current_command_context() -> None:
    result = _run_cli("context-direct")

    assert result.returncode == 0, result.stderr
    assert result.stdout.strip() == "context-direct:same"


def test_context_service_returns_current_command_context() -> None:
    result = _run_cli("context-service")

    assert result.returncode == 0, result.stderr
    assert result.stdout.strip() == "context-service:same"



"""Two readability-focused patterns.

1. ``decorate(...)`` can be called before ``add_*``.
2. ``inner_parameter=...`` removes ambiguity when inference is unclear.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Annotated

from diwire import Component, Container
from diwire.exceptions import DIWireInvalidRegistrationError


class Repo:
    def get(self, key: str) -> str:
        raise NotImplementedError


class SqlRepo(Repo):
    def get(self, key: str) -> str:
        return f"sql:{key}"


PrimaryRepo = Annotated[Repo, Component("primary")]


@dataclass(slots=True)
class CachedRepo(Repo):
    inner: Repo
    cache_hits: int = field(default=0, init=False)

    def get(self, key: str) -> str:
        self.cache_hits += 1
        return self.inner.get(key)


class AmbiguousDecorator(Repo):
    def __init__(self, first: Repo, second: Repo) -> None:
        self.first = first
        self.second = second

    def get(self, key: str) -> str:
        return self.first.get(key)


def main() -> None:
    # Pattern A: decorate first, register later.
    container = Container()
    container.decorate(
        provides=PrimaryRepo,
        decorator=CachedRepo,
        inner_parameter="inner",
    )
    container.add(SqlRepo, provides=PrimaryRepo)

    decorated = container.resolve(PrimaryRepo)
    print(f"pattern_a_outer={type(decorated).__name__}")  # => pattern_a_outer=CachedRepo
    print(f"pattern_a_inner={type(decorated.inner).__name__}")  # => pattern_a_inner=SqlRepo
    print(f"pattern_a_result={decorated.get('account-42')}")  # => pattern_a_result=sql:account-42

    # Pattern B: ambiguous decorator needs explicit inner_parameter.
    ambiguous_error: str
    try:
        container.decorate(provides=Repo, decorator=AmbiguousDecorator)
    except DIWireInvalidRegistrationError as error:
        ambiguous_error = type(error).__name__
    print(f"ambiguous_error={ambiguous_error}")  # => ambiguous_error=DIWireInvalidRegistrationError

    container.decorate(
        provides=Repo,
        decorator=AmbiguousDecorator,
        inner_parameter="first",
    )
    print(f"pattern_b_error={ambiguous_error}")  # => pattern_b_error=DIWireInvalidRegistrationError
    print(
        "pattern_b_inner_parameter_accepts_registration=True"
    )  # => pattern_b_inner_parameter_accepts_registration=True


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import AsyncIterator, Awaitable, Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, cast

import pytest

from diwire import Container, Injected, Lifetime, Scope, resolver_context

if TYPE_CHECKING:
    from aiohttp import web
    from aiohttp.test_utils import TestClient


@dataclass
class _RequestPathService:
    request: web.Request

    def path(self) -> str:
        return self.request.path


@dataclass
class _WebSocketPathService:
    request: web.Request

    def path(self) -> str:
        return self.request.path


@pytest.fixture()
def app() -> web.Application:
    pytest.importorskip("aiohttp")

    from aiohttp import web

    from diwire.integrations.aiohttp import add_request_context, request_context_middleware

    # diwire resolves dependencies via runtime type hints; populate module globals so
    # forward references like `web.Request` can be evaluated.
    globals()["web"] = web

    container = Container()
    add_request_context(container)
    container.add(
        _RequestPathService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _WebSocketPathService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    app = web.Application(middlewares=[request_context_middleware])

    @resolver_context.inject(scope=Scope.REQUEST)
    async def request_direct(
        request: web.Request,
        resolved_request: Injected[web.Request],
    ) -> web.StreamResponse:
        return web.json_response({"path": resolved_request.path})

    @resolver_context.inject(scope=Scope.REQUEST)
    async def request_service(
        request: web.Request,
        service: Injected[_RequestPathService],
    ) -> web.StreamResponse:
        return web.json_response({"path": service.path()})

    @resolver_context.inject(scope=Scope.REQUEST)
    async def websocket_direct(
        request: web.Request,
        resolved_request: Injected[web.Request],
    ) -> web.StreamResponse:
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        await ws.send_json({"path": resolved_request.path})
        await ws.close()
        return ws

    @resolver_context.inject(scope=Scope.REQUEST)
    async def websocket_service(
        request: web.Request,
        service: Injected[_WebSocketPathService],
    ) -> web.StreamResponse:
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        await ws.send_json({"path": service.path()})
        await ws.close()
        return ws

    app.add_routes(
        [
            web.get(
                "/request/direct",
                cast("Callable[[web.Request], Awaitable[web.StreamResponse]]", request_direct),
            ),
            web.get(
                "/request/service",
                cast("Callable[[web.Request], Awaitable[web.StreamResponse]]", request_service),
            ),
            web.get(
                "/websocket/direct",
                cast("Callable[[web.Request], Awaitable[web.StreamResponse]]", websocket_direct),
            ),
            web.get(
                "/websocket/service",
                cast("Callable[[web.Request], Awaitable[web.StreamResponse]]", websocket_service),
            ),
        ]
    )
    return app


@pytest.fixture()
async def client(app: web.Application) -> AsyncIterator[TestClient[web.Request, web.Application]]:
    pytest.importorskip("aiohttp")

    from aiohttp.test_utils import TestClient, TestServer

    async with TestServer(app) as server:
        async with TestClient(server) as test_client:
            yield test_client


@pytest.mark.asyncio
async def test_request_resolve_for_http_endpoint(
    client: TestClient[web.Request, web.Application],
) -> None:
    response = await client.get("/request/direct")
    assert response.status == 200
    assert await response.json() == {"path": "/request/direct"}


@pytest.mark.asyncio
async def test_request_resolve_in_service_for_http_endpoint(
    client: TestClient[web.Request, web.Application],
) -> None:
    response = await client.get("/request/service")
    assert response.status == 200
    assert await response.json() == {"path": "/request/service"}


@pytest.mark.asyncio
async def test_websocket_resolve_for_websocket_endpoint(
    client: TestClient[web.Request, web.Application],
) -> None:
    websocket = await client.ws_connect("/websocket/direct")
    payload = await websocket.receive_json()
    assert payload == {"path": "/websocket/direct"}
    await websocket.close()


@pytest.mark.asyncio
async def test_websocket_resolve_in_service_for_websocket_endpoint(
    client: TestClient[web.Request, web.Application],
) -> None:
    websocket = await client.ws_connect("/websocket/service")
    payload = await websocket.receive_json()
    assert payload == {"path": "/websocket/service"}
    await websocket.close()



from __future__ import annotations

import pytest

pytest.importorskip("flask")

from flask import Flask, Request

import diwire._internal.integrations.flask as flask_integration
from diwire import Container
from diwire.exceptions import DIWireIntegrationError


def test_get_request_raises_when_context_is_missing() -> None:
    with pytest.raises(DIWireIntegrationError, match="Request context not available"):
        flask_integration.get_request()


def test_get_request_returns_current_flask_request() -> None:
    app = Flask(__name__)
    with app.test_request_context("/middleware/context"):
        resolved_request = flask_integration.get_request()
        assert resolved_request.path == "/middleware/context"


def test_add_request_context_registers_factory_for_container_resolution() -> None:
    container = Container()
    flask_integration.add_request_context(container)

    app = Flask(__name__)
    with app.test_request_context("/container/request"):
        resolved_request = container.resolve(Request)
        assert resolved_request.path == "/container/request"



from collections.abc import Iterator
from dataclasses import dataclass
from json import loads
from typing import cast

import pytest

pytest.importorskip("django")
pytest.importorskip("dmr")
pytest.importorskip("pydantic")

from django.http import HttpRequest, HttpResponse
from django.test import override_settings
from django.urls import path
from dmr import Controller
from dmr.plugins.pydantic import PydanticSerializer
from dmr.test import DMRClient
from pydantic import BaseModel

from diwire import Container, Injected, Lifetime, Scope, resolver_context
from diwire.integrations.django import add_request_context
from tests._django_setup import ensure_django_setup

ensure_django_setup()


def _json_body(response: HttpResponse) -> dict[str, str]:
    content = cast("bytes", response.content)
    return cast("dict[str, str]", loads(content))


class _PathResponse(BaseModel):
    path: str


@dataclass
class _RequestPathService:
    request: HttpRequest

    def path(self) -> str:
        return self.request.path


_container = Container()
add_request_context(_container)
_container.add(
    _RequestPathService,
    scope=Scope.REQUEST,
    lifetime=Lifetime.SCOPED,
)


class _DirectRequestController(Controller[PydanticSerializer]):
    @resolver_context.inject(scope=Scope.REQUEST)
    def get(self, request: Injected[HttpRequest]) -> _PathResponse:
        return _PathResponse(path=request.path)


class _RequestServiceController(Controller[PydanticSerializer]):
    @resolver_context.inject(scope=Scope.REQUEST)
    def get(self, service: Injected[_RequestPathService]) -> _PathResponse:
        return _PathResponse(path=service.path())


urlpatterns = [
    path("request/direct/", _DirectRequestController.as_view()),
    path("request/service/", _RequestServiceController.as_view()),
]


@pytest.fixture()
def client() -> Iterator[DMRClient]:
    with override_settings(
        ROOT_URLCONF=__name__,
        MIDDLEWARE=["diwire.integrations.django.RequestContextMiddleware"],
    ):
        yield DMRClient()


def test_request_resolve_for_controller_endpoint(client: DMRClient) -> None:
    response = client.get("/request/direct/")
    assert response.status_code == 200
    assert _json_body(response) == {"path": "/request/direct/"}


def test_request_resolve_in_service_for_controller_endpoint(client: DMRClient) -> None:
    response = client.get("/request/service/")
    assert response.status_code == 200
    assert _json_body(response) == {"path": "/request/service/"}



from __future__ import annotations

from typing import cast

import pytest

pytest.importorskip("click")
pytest.importorskip("typer")

import click
import typer

import diwire._internal.integrations.typer as typer_integration
from diwire import Container
from diwire.exceptions import DIWireIntegrationError


def test_get_typer_context_raises_when_context_is_missing() -> None:
    with pytest.raises(DIWireIntegrationError, match="Typer context not available"):
        typer_integration.get_typer_context()


def test_get_typer_context_returns_current_context() -> None:
    with click.Context(click.Command("context-direct"), info_name="context-direct") as expected:
        resolved = typer_integration.get_typer_context()

    assert resolved is expected


def test_add_typer_context_registers_factory_for_container_resolution() -> None:
    container = Container()
    typer_integration.add_typer_context(container)

    with click.Context(click.Command("context-service"), info_name="context-service"):
        expected = cast("typer.Context", click.get_current_context())
        resolved = container.resolve(typer.Context)

    assert resolved is expected



"""Focused example: ``add_instance`` for pre-built objects."""

from __future__ import annotations

from dataclasses import dataclass

from diwire import Container


@dataclass(slots=True)
class Config:
    value: str


def main() -> None:
    container = Container()
    config = Config(value="singleton")
    container.add_instance(config, provides=Config)

    first = container.resolve(Config)
    second = container.resolve(Config)
    print(f"instance_singleton={first is second}")  # => instance_singleton=True


if __name__ == "__main__":
    main()



from pydantic import Field
from pydantic_settings import BaseSettings


class AioHttpE2ESettings(BaseSettings):
    host: str = Field(default="localhost", alias="AIOHTTP_E2E_HOST")
    port: int = Field(default=8230, alias="AIOHTTP_E2E_PORT")



"""Tiny, dependency-free sitemap generator for SEO.

This intentionally stays minimal (and vendored) to avoid pulling extra docs
dependencies into the library.
"""

from __future__ import annotations

import contextlib
import html
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Protocol

from sphinx.errors import ExtensionError


class _SphinxConfig(Protocol):
    html_baseurl: str


class _SphinxBuilder(Protocol):
    format: str


class _SphinxApp(Protocol):
    builder: _SphinxBuilder
    config: _SphinxConfig
    outdir: str

    def add_config_value(self, name: str, default: Any, rebuild: str) -> None: ...

    def connect(self, event: str, callback: Any) -> None: ...


@dataclass(frozen=True, slots=True)
class _SitemapEntry:
    loc: str
    lastmod: str


def _canonical_path_from_html_rel(rel: str) -> str:
    if rel == "index.html":
        return "/"
    if rel.endswith("/index.html"):
        return f"/{rel.removesuffix('index.html')}"
    return f"/{rel.removesuffix('.html')}/"


def _is_generated_redirect_index(rel: str, *, outdir: Path) -> bool:
    if not rel.endswith("/index.html"):
        return False

    non_index_rel = f"{rel.removesuffix('/index.html')}.html"
    return (outdir / non_index_rel).is_file()


def _iter_html_files(outdir: Path) -> list[Path]:
    # Sphinx outputs a lot of internal files; we only want real pages.
    html_files: list[Path] = []
    for path in outdir.rglob("*.html"):
        rel = path.relative_to(outdir).as_posix()
        if rel.startswith(("_static/", "_sources/", "_modules/")):
            continue
        if rel in {"genindex.html", "search.html"}:
            continue
        if _is_generated_redirect_index(rel, outdir=outdir):
            continue
        html_files.append(path)

    # Stable ordering helps keep diffs small (and makes local debugging nicer).
    html_files.sort(key=lambda p: p.relative_to(outdir).as_posix())
    return html_files


def _build_entries(outdir: Path, *, baseurl: str) -> list[_SitemapEntry]:
    base = baseurl.rstrip("/")
    entries: list[_SitemapEntry] = []
    for html_path in _iter_html_files(outdir):
        rel = html_path.relative_to(outdir).as_posix()
        loc = f"{base}{_canonical_path_from_html_rel(rel)}"
        lastmod = (
            datetime.fromtimestamp(html_path.stat().st_mtime, tz=timezone.utc).date().isoformat()
        )
        entries.append(_SitemapEntry(loc=loc, lastmod=lastmod))
    return entries


def _write_sitemap(outdir: Path, entries: list[_SitemapEntry]) -> None:
    # Keep XML output stable and small; don't add unnecessary optional tags.
    lines = [
        '<?xml version="1.0" encoding="utf-8"?>',
        '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">',
    ]
    for entry in entries:
        lines.extend(
            [
                "  <url>",
                f"    <loc>{entry.loc}</loc>",
                f"    <lastmod>{entry.lastmod}</lastmod>",
                "  </url>",
            ],
        )
    lines.append("</urlset>")

    (outdir / "sitemap.xml").write_text("\n".join(lines) + "\n", encoding="utf-8")


def _write_robots(outdir: Path, *, baseurl: str) -> None:
    base = baseurl.rstrip("/") + "/"
    content = "\n".join(
        [
            "User-agent: *",
            "Allow: /",
            "",
            f"Sitemap: {base}sitemap.xml",
            "",
        ],
    )
    (outdir / "robots.txt").write_text(content, encoding="utf-8")


def _render_redirect_page(*, target_href: str, canonical_url: str) -> str:
    escaped_target = html.escape(target_href, quote=True)
    escaped_canonical = html.escape(canonical_url, quote=True)
    return "\n".join(
        [
            "<!doctype html>",
            '<html lang="en">',
            "  <head>",
            '    <meta charset="utf-8">',
            "    <title>Redirecting...</title>",
            f'    <link rel="canonical" href="{escaped_canonical}">',
            f'    <meta http-equiv="refresh" content="0; url={escaped_target}">',
            f'    <script>window.location.replace("{escaped_target}");</script>',
            "  </head>",
            "  <body>",
            f'    <p>Redirecting to <a href="{escaped_target}">{escaped_target}</a>.</p>',
            "  </body>",
            "</html>",
            "",
        ],
    )


def _write_extensionless_redirects(outdir: Path, *, html_files: list[Path], baseurl: str) -> None:
    base = baseurl.rstrip("/")
    for html_path in html_files:
        rel = html_path.relative_to(outdir).as_posix()
        if rel.endswith("index.html"):
            continue

        source_name = Path(rel).name
        redirect_rel = Path(rel).with_suffix("") / "index.html"
        redirect_path = outdir / redirect_rel
        redirect_path.parent.mkdir(parents=True, exist_ok=True)

        canonical_url = f"{base}{_canonical_path_from_html_rel(rel)}"
        redirect_html = _render_redirect_page(
            target_href=f"../{source_name}",
            canonical_url=canonical_url,
        )
        redirect_path.write_text(redirect_html, encoding="utf-8")


def _build_finished(app: _SphinxApp, exc: Exception | None) -> None:
    if exc is not None:
        return
    if app.builder.format != "html":
        return

    baseurl = app.config.html_baseurl.strip()
    if not baseurl:
        return

    outdir = Path(app.outdir)
    html_files = _iter_html_files(outdir)
    if not html_files:
        return

    _write_extensionless_redirects(outdir, html_files=html_files, baseurl=baseurl)
    entries = _build_entries(outdir, baseurl=baseurl)
    if not entries:
        return

    _write_sitemap(outdir, entries)
    _write_robots(outdir, baseurl=baseurl)


def setup(app: _SphinxApp) -> dict[str, Any]:
    # We use the built-in `html_baseurl` config as the canonical site URL.
    # Some Sphinx versions predefine it; treat that as OK.
    with contextlib.suppress(ExtensionError):
        app.add_config_value("html_baseurl", default="", rebuild="html")
    app.connect("build-finished", _build_finished)

    return {
        "version": "0.1.0",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from __future__ import annotations

import functools
import inspect
from collections.abc import Awaitable, Callable
from contextvars import ContextVar, Token
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Any, Literal, TypeVar, cast, overload

from diwire._internal.injection import (
    INJECT_RESOLVER_KWARG,
    INJECT_WRAPPER_MARKER,
    InjectedCallableInspector,
)
from diwire._internal.policies import DependencyRegistrationPolicy
from diwire._internal.resolvers.protocol import ResolverProtocol
from diwire._internal.scope import BaseScope
from diwire.exceptions import DIWireInvalidRegistrationError, DIWireResolverNotSetError

if TYPE_CHECKING:
    from typing import Protocol

    from typing_extensions import Self

    from diwire._internal.container import Container


T = TypeVar("T")
InjectReturnT_co = TypeVar("InjectReturnT_co", covariant=True)

if TYPE_CHECKING:

    class _InjectedCallable(Protocol[InjectReturnT_co]):
        def __call__(
            self,
            *args: Any,
            diwire_resolver: ResolverProtocol = ...,
            **kwargs: Any,
        ) -> InjectReturnT_co: ...

    class _InjectDecorator(Protocol):
        def __call__(
            self,
            func: Callable[..., InjectReturnT_co],
            /,
        ) -> _InjectedCallable[InjectReturnT_co]: ...


@dataclass(frozen=True, slots=True)
class _InjectInvocationState:
    source: Literal["explicit", "context", "fallback"]
    context_resolver: ResolverProtocol | None


@dataclass(frozen=True, slots=True)
class _InjectWrapperConfig:
    callable_obj: Callable[..., Any]
    scope: BaseScope | None
    dependency_registration_policy: DependencyRegistrationPolicy | None
    auto_open_scope: bool


class _ResolverBoundResolver:
    """Resolver wrapper that synchronizes resolver context with ResolverContext."""

    def __init__(
        self,
        *,
        resolver: ResolverProtocol,
        resolver_context: ResolverContext,
        push_resolver: Callable[[ResolverProtocol], None],
        pop_resolver: Callable[[], None],
    ) -> None:
        self._resolver = resolver
        self._resolver_context = resolver_context
        self._push_resolver = push_resolver
        self._pop_resolver = pop_resolver

    def __getattr__(self, name: str) -> Any:
        return getattr(self._resolver, name)

    @overload
    def resolve(self, dependency: type[T]) -> T: ...

    @overload
    def resolve(self, dependency: Any) -> Any: ...

    def resolve(self, dependency: Any) -> Any:
        return self._resolver.resolve(dependency)

    @overload
    async def aresolve(self, dependency: type[T]) -> T: ...

    @overload
    async def aresolve(self, dependency: Any) -> Any: ...

    async def aresolve(self, dependency: Any) -> Any:
        return await self._resolver.aresolve(dependency)

    def enter_scope(
        self,
        scope: BaseScope | None = None,
    ) -> _ResolverBoundResolver:
        scoped_resolver = self._resolver.enter_scope(scope)
        if scoped_resolver is self._resolver:
            return self
        return _ResolverBoundResolver(
            resolver=scoped_resolver,
            resolver_context=self._resolver_context,
            push_resolver=self._push_resolver,
            pop_resolver=self._pop_resolver,
        )

    def __enter__(self) -> Self:
        self._resolver.__enter__()
        self._push_resolver(cast("ResolverProtocol", self))
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        try:
            self._resolver.__exit__(exc_type, exc_value, traceback)
        finally:
            self._pop_resolver()

    async def __aenter__(self) -> Self:
        await cast("Any", self._resolver).__aenter__()
        self._push_resolver(cast("ResolverProtocol", self))
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        try:
            await self._resolver.__aexit__(exc_type, exc_value, traceback)
        finally:
            self._pop_resolver()

    def close(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self._resolver.close(exc_type, exc_value, traceback)

    async def aclose(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self._resolver.aclose(exc_type, exc_value, traceback)


class ResolverContext:
    """Task/thread-safe context for resolver-bound injection and resolution."""

    __slots__ = (
        "_current_resolver_var",
        "_fallback_container",
        "_injected_callable_inspector",
        "_token_stack_var",
    )

    def __init__(self) -> None:
        self._current_resolver_var: ContextVar[ResolverProtocol | None] = ContextVar(
            "diwire_resolver_context_resolver",
            default=None,
        )
        self._token_stack_var: ContextVar[tuple[Token[ResolverProtocol | None], ...]] = ContextVar(
            "diwire_resolver_context_tokens",
            default=(),
        )
        self._fallback_container: Container | None = None
        self._injected_callable_inspector = InjectedCallableInspector()

    def set_fallback_container(self, container: Container) -> None:
        """Set the fallback container used when no resolver is bound.

        Resolver-bound contexts always take precedence over this fallback.

        Args:
            container: Container used to compile fallback resolvers and injected
                callables when no context resolver is active.

        """
        self._set_fallback_container(container)

    def _push(self, resolver: ResolverProtocol) -> None:
        token = self._current_resolver_var.set(resolver)
        tokens = self._token_stack_var.get()
        self._token_stack_var.set((*tokens, token))

    def _pop(self) -> None:
        tokens = self._token_stack_var.get()
        if not tokens:
            return
        token = tokens[-1]
        self._token_stack_var.set(tokens[:-1])
        self._current_resolver_var.reset(token)

    def _set_fallback_container(self, container: Container) -> None:
        self._fallback_container = container

    def _get_bound_resolver_or_none(self) -> ResolverProtocol | None:
        return self._current_resolver_var.get()

    def _require_context_or_fallback_resolver(self) -> ResolverProtocol:
        resolver = self._get_bound_resolver_or_none()
        if resolver is not None:
            return resolver

        fallback_resolver = self._get_fallback_resolver_or_none()
        if fallback_resolver is not None:
            return fallback_resolver

        msg = (
            "Resolver is not set for resolver_context. Enter a compiled resolver context "
            "before using resolver_context."
        )
        raise DIWireResolverNotSetError(msg)

    def _get_fallback_resolver_or_none(self) -> ResolverProtocol | None:
        fallback_container = self._fallback_container
        if fallback_container is None:
            return None

        fallback_resolver = fallback_container.compile()
        return self._wrap_resolver(fallback_resolver)

    def _wrap_resolver(self, resolver: ResolverProtocol) -> ResolverProtocol:
        resolver_any = cast("Any", resolver)
        if isinstance(resolver_any, _ResolverBoundResolver):
            return cast("ResolverProtocol", resolver_any)
        return cast(
            "ResolverProtocol",
            _ResolverBoundResolver(
                resolver=resolver,
                resolver_context=self,
                push_resolver=self._push,
                pop_resolver=self._pop,
            ),
        )

    @overload
    def resolve(self, dependency: type[T]) -> T: ...

    @overload
    def resolve(self, dependency: Any) -> Any: ...

    def resolve(self, dependency: Any) -> Any:
        """Resolve a dependency from the active resolver or fallback container.

        Args:
            dependency: Dependency key to resolve.

        Raises:
            DIWireResolverNotSetError: If no resolver is bound and no fallback
                container is configured.

        """
        return self._require_context_or_fallback_resolver().resolve(dependency)

    @overload
    async def aresolve(self, dependency: type[T]) -> T: ...

    @overload
    async def aresolve(self, dependency: Any) -> Any: ...

    async def aresolve(self, dependency: Any) -> Any:
        """Asynchronously resolve a dependency from context or fallback.

        Args:
            dependency: Dependency key to resolve.

        Raises:
            DIWireResolverNotSetError: If no resolver is bound and no fallback
                container is configured.

        """
        return await self._require_context_or_fallback_resolver().aresolve(dependency)

    def enter_scope(
        self,
        scope: BaseScope | None = None,
    ) -> ResolverProtocol:
        """Enter a child scope on the active resolver or fallback resolver.

        Args:
            scope: Target scope to enter. ``None`` delegates to
                ``_require_context_or_fallback_resolver().enter_scope(None)``,
                so behavior follows the resolver's default transition semantics
                (see ``ResolverProtocol.enter_scope``).

        Raises:
            DIWireResolverNotSetError: If no resolver is bound and no fallback
                container is configured.

        """
        return self._require_context_or_fallback_resolver().enter_scope(scope)

    @overload
    def inject(
        self,
        func: Callable[..., InjectReturnT_co],
    ) -> _InjectedCallable[InjectReturnT_co]: ...

    @overload
    def inject(
        self,
        func: Literal["from_decorator"] = "from_decorator",
        *,
        scope: BaseScope | Literal["infer"] = "infer",
        dependency_registration_policy: (
            DependencyRegistrationPolicy | Literal["from_container"]
        ) = "from_container",
        auto_open_scope: bool = True,
    ) -> _InjectDecorator: ...

    def inject(
        self,
        func: Callable[..., Any] | Literal["from_decorator"] = "from_decorator",
        *,
        scope: BaseScope | Literal["infer"] = "infer",
        dependency_registration_policy: (
            DependencyRegistrationPolicy | Literal["from_container"]
        ) = "from_container",
        auto_open_scope: bool = True,
    ) -> _InjectedCallable[Any] | _InjectDecorator:
        """Wrap callables so ``Injected[...]`` parameters resolve at invocation.

        Resolution precedence at call time is explicit ``diwire_resolver``,
        then an active context-bound resolver, then the configured fallback
        container when it is configured with ``use_resolver_context=True``.

        Args:
            func: Callable to wrap directly, or ``"from_decorator"`` when used
                as ``@resolver_context.inject(...)``.
            scope: Explicit scope for wrapper generation, or ``"infer"`` to
                infer from injected dependencies.
            dependency_registration_policy: Dependency autoregistration policy for
                wrapper generation, or ``"from_container"`` to inherit the
                fallback container setting.
            auto_open_scope: Whether invocation should auto-enter scopes when
                needed. When ``True`` (default), scope entry is attempted only
                when moving into a deeper target scope is valid and required.
                If the target scope is already open, no additional scope is
                entered. If the current resolver is already deeper than the
                target scope, no additional scope is entered and resolution
                proceeds from the current resolver.

        Raises:
            DIWireInvalidRegistrationError: If inject configuration values are
                invalid, ``func`` is not callable, or the callable uses
                reserved parameter names.
            DIWireResolverNotSetError: If invocation has no explicit resolver,
                no active context resolver, and no fallback container eligible
                for inject fallback.

        """
        resolved_scope = self._resolve_inject_scope(scope)
        resolved_dependency_registration_policy = (
            self._resolve_inject_dependency_registration_policy(
                dependency_registration_policy=dependency_registration_policy,
            )
        )

        def decorator(
            callable_obj: Callable[..., InjectReturnT_co],
        ) -> _InjectedCallable[InjectReturnT_co]:
            self._validate_injected_callable_signature(callable_obj)
            inspected_callable = self._injected_callable_inspector.inspect_callable(callable_obj)
            cache: dict[Container, Callable[..., Any]] = {}
            wrapper_config = _InjectWrapperConfig(
                callable_obj=callable_obj,
                scope=resolved_scope,
                dependency_registration_policy=resolved_dependency_registration_policy,
                auto_open_scope=auto_open_scope,
            )
            fallback_container = self._fallback_container
            if fallback_container is not None:
                self._get_cached_injected_callable(
                    cache=cache,
                    container=fallback_container,
                    wrapper_config=wrapper_config,
                )

            invocation = self._build_injection_invoker(
                cache=cache,
                wrapper_config=wrapper_config,
            )
            wrapped_callable = self._wrap_injected_callable(
                callable_obj=callable_obj,
                invocation=invocation,
            )
            wrapped_callable.__signature__ = inspected_callable.public_signature  # type: ignore[attr-defined]
            wrapped_callable.__dict__[INJECT_WRAPPER_MARKER] = True
            return cast("_InjectedCallable[InjectReturnT_co]", wrapped_callable)

        func_value = cast("Any", func)
        if func_value == "from_decorator":
            return cast("_InjectDecorator", decorator)
        if not callable(func_value):
            msg = "inject() parameter 'func' must be callable or 'from_decorator'."
            raise DIWireInvalidRegistrationError(msg)
        return decorator(func_value)

    def _resolve_inject_scope(
        self,
        scope: BaseScope | Literal["infer"],
    ) -> BaseScope | None:
        scope_value = cast("Any", scope)
        if scope_value == "infer":
            return None
        if isinstance(scope_value, BaseScope):
            return scope_value
        msg = "inject() parameter 'scope' must be BaseScope or 'infer'."
        raise DIWireInvalidRegistrationError(msg)

    def _resolve_inject_dependency_registration_policy(
        self,
        *,
        dependency_registration_policy: (DependencyRegistrationPolicy | Literal["from_container"]),
    ) -> DependencyRegistrationPolicy | None:
        dependency_registration_policy_value = cast("Any", dependency_registration_policy)
        if dependency_registration_policy_value == "from_container":
            return None
        if isinstance(dependency_registration_policy_value, DependencyRegistrationPolicy):
            return dependency_registration_policy_value
        msg = (
            "inject() parameter 'dependency_registration_policy' must be "
            "DependencyRegistrationPolicy or 'from_container'."
        )
        raise DIWireInvalidRegistrationError(msg)

    def _validate_injected_callable_signature(self, callable_obj: Callable[..., Any]) -> None:
        signature = inspect.signature(callable_obj)
        if INJECT_RESOLVER_KWARG in signature.parameters:
            msg = (
                f"Callable '{self._callable_name(callable_obj)}' cannot declare reserved "
                f"parameter '{INJECT_RESOLVER_KWARG}'."
            )
            raise DIWireInvalidRegistrationError(msg)

    def _build_injection_invoker(
        self,
        *,
        cache: dict[Container, Callable[..., Any]],
        wrapper_config: _InjectWrapperConfig,
    ) -> Callable[..., Any]:
        def _invoke(*args: Any, **kwargs: Any) -> Any:
            state = self._resolve_injection_state(kwargs)
            fallback_container = self._require_inject_fallback_container()
            injected_callable = self._get_cached_injected_callable(
                cache=cache,
                container=fallback_container,
                wrapper_config=wrapper_config,
            )

            runtime_kwargs = kwargs
            if state.source == "context":
                runtime_kwargs = dict(kwargs)
                runtime_kwargs[INJECT_RESOLVER_KWARG] = cast(
                    "ResolverProtocol",
                    state.context_resolver,
                )
            return injected_callable(*args, **runtime_kwargs)

        return _invoke

    def _resolve_injection_state(self, kwargs: dict[str, Any]) -> _InjectInvocationState:
        if INJECT_RESOLVER_KWARG in kwargs:
            return _InjectInvocationState(source="explicit", context_resolver=None)

        context_resolver = self._get_bound_resolver_or_none()
        if context_resolver is not None:
            return _InjectInvocationState(source="context", context_resolver=context_resolver)

        fallback_container = self._fallback_container
        if fallback_container is not None:
            fallback_uses_resolver_context = bool(
                getattr(fallback_container, "_use_resolver_context", True)
            )
            if fallback_uses_resolver_context:
                fallback_container.compile()
                return _InjectInvocationState(source="fallback", context_resolver=None)

        msg = (
            "Resolver is not set for resolver_context.inject. Pass "
            f"'{INJECT_RESOLVER_KWARG}' explicitly, enter a resolver context, "
            "or initialize a fallback container with use_resolver_context=True."
        )
        raise DIWireResolverNotSetError(msg)

    def _require_inject_fallback_container(self) -> Container:
        fallback_container = self._fallback_container
        if fallback_container is None:
            msg = (
                "ResolverContext.inject requires a fallback container. Initialize a container "
                "with this ResolverContext before decorating callables."
            )
            raise DIWireResolverNotSetError(msg)
        return fallback_container

    def _get_cached_injected_callable(
        self,
        *,
        cache: dict[Container, Callable[..., Any]],
        container: Container,
        wrapper_config: _InjectWrapperConfig,
    ) -> Callable[..., Any]:
        injected_callable = cache.get(container)
        if injected_callable is not None:
            return injected_callable

        injected_callable = container._inject_callable(  # noqa: SLF001
            callable_obj=wrapper_config.callable_obj,
            scope=wrapper_config.scope,
            dependency_registration_policy=wrapper_config.dependency_registration_policy,
            auto_open_scope=wrapper_config.auto_open_scope,
        )
        cache[container] = injected_callable
        return injected_callable

    def _wrap_injected_callable(
        self,
        *,
        callable_obj: Callable[..., Any],
        invocation: Callable[..., Any],
    ) -> Callable[..., Any]:
        if inspect.iscoroutinefunction(callable_obj):

            @functools.wraps(callable_obj)
            async def _async_injected(*args: Any, **kwargs: Any) -> Any:
                result = invocation(*args, **kwargs)
                return await cast("Awaitable[Any]", result)

            return _async_injected

        @functools.wraps(callable_obj)
        def _sync_injected(*args: Any, **kwargs: Any) -> Any:
            return invocation(*args, **kwargs)

        return _sync_injected

    def _callable_name(self, callable_obj: Callable[..., Any]) -> str:
        return getattr(callable_obj, "__qualname__", repr(callable_obj))


resolver_context = ResolverContext()



"""Focused example: caller override for injected parameters."""

from __future__ import annotations

from dataclasses import dataclass

from diwire import Container, Injected, resolver_context


@dataclass(slots=True)
class User:
    email: str


def main() -> None:
    container = Container()
    container.add_instance(User(email="container@example.com"))

    @resolver_context.inject
    def handler(user: Injected[User]) -> str:
        return user.email

    default_value = handler()
    override_value = handler(user=User(email="override@example.com"))

    print(f"default_injected={default_value}")  # => default_injected=container@example.com
    print(f"override_injected={override_value}")  # => override_injected=override@example.com


if __name__ == "__main__":
    main()



from __future__ import annotations

from contextvars import ContextVar
from typing import Any, cast

from starlette.requests import HTTPConnection, Request, StateT
from starlette.types import ASGIApp, Receive, Scope, Send
from starlette.websockets import WebSocket

from diwire import Container, Lifetime
from diwire.exceptions import DIWireIntegrationError

_connection_context: ContextVar[HTTPConnection | None] = ContextVar(
    "diwire_fastapi_connection_context",
    default=None,
)


def get_connection(_state_t: type[StateT]) -> HTTPConnection[StateT]:
    connection = _connection_context.get()
    if connection is None:
        msg = "Connection context not available. Ensure RequestContextMiddleware is installed."
        raise DIWireIntegrationError(msg)
    return cast("HTTPConnection[StateT]", connection)


def get_request(_state_t: type[StateT]) -> Request[StateT]:
    connection = get_connection(_state_t)
    if not isinstance(connection, Request):
        msg = "Current connection is not HTTP Request."
        raise DIWireIntegrationError(msg)
    return cast("Request[StateT]", connection)


def _get_websocket_typed(_state_t: type[StateT]) -> WebSocket[StateT]:  # type: ignore[type-arg]
    connection = get_connection(_state_t)
    if not isinstance(connection, WebSocket):
        msg = "Current connection is not WebSocket."
        raise DIWireIntegrationError(msg)
    return cast("WebSocket[StateT]", connection)  # type: ignore[redundant-cast, type-arg]


def _get_websocket_untyped() -> WebSocket:
    connection = get_connection(Any)  # type: ignore[type-var]
    if not isinstance(connection, WebSocket):  # type: ignore[unreachable]
        msg = "Current connection is not WebSocket."
        raise DIWireIntegrationError(msg)
    return connection  # type: ignore[unreachable]


try:
    if StateT in WebSocket.__orig_bases__[0].__args__:  # type: ignore[attr-defined]
        get_websocket = _get_websocket_typed
    else:
        get_websocket = _get_websocket_untyped  # type: ignore[assignment]
except (AttributeError, IndexError, TypeError):
    # WebSocket does not have __orig_bases__ or is not generic, fallback to untyped version
    get_websocket = _get_websocket_untyped  # type: ignore[assignment]


class RequestContextMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        scope_type = scope["type"]

        if scope_type == "http":
            connection: HTTPConnection = Request(scope, receive=receive)
        elif scope_type == "websocket":
            connection = WebSocket(scope, receive=receive, send=send)
        else:
            await self.app(scope, receive, send)
            return

        token = _connection_context.set(connection)

        try:
            await self.app(scope, receive, send)
        finally:
            _connection_context.reset(token)


def add_request_context(container: Container) -> None:
    container.add_factory(
        get_connection,
        provides=HTTPConnection,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add_factory(
        get_request,
        provides=Request,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add_factory(
        get_websocket,
        provides=WebSocket,
        lifetime=Lifetime.TRANSIENT,
    )



from collections.abc import Awaitable, Callable
from typing import cast

from aiohttp import WSMsgType, web

from diwire import Container, Injected, Scope, resolver_context
from diwire.integrations.aiohttp import add_request_context, request_context_middleware
from tests.e2e.aiohttp.setup.config import AioHttpE2ESettings
from tests.e2e.aiohttp.setup.services import (
    CMService,
    RequestBasedService,
    WebSocketBasedService,
    latest_cleanup_path,
)

app = web.Application(middlewares=[request_context_middleware])


async def health(_request: web.Request) -> web.StreamResponse:
    return web.Response(text="OK")


@resolver_context.inject(scope=Scope.REQUEST)
async def request_based_service(
    request: web.Request,
    service: Injected[RequestBasedService],
) -> web.StreamResponse:
    return web.Response(text=service.work())


@resolver_context.inject(scope=Scope.REQUEST)
async def cm_service(
    request: web.Request,
    service: Injected[CMService],
) -> web.StreamResponse:
    return web.Response(text=service.work())


async def cm_service_cleanup(_request: web.Request) -> web.StreamResponse:
    cleanup_path = latest_cleanup_path()
    return web.Response(text=f"Cleanup path: {cleanup_path}")


@resolver_context.inject(scope=Scope.REQUEST)
async def websocket_request_based_service(
    request: web.Request,
    service: Injected[WebSocketBasedService],
) -> web.StreamResponse:
    ws = web.WebSocketResponse()
    await ws.prepare(request)
    async for message in ws:
        if message.type == WSMsgType.TEXT:
            await ws.send_str(service.work())
            break
    await ws.close()
    return ws


app.add_routes(
    [
        web.get("/health", health),
        web.get(
            "/services/request-based",
            cast(
                "Callable[[web.Request], Awaitable[web.StreamResponse]]",
                request_based_service,
            ),
        ),
        web.get(
            "/services/cm-service",
            cast(
                "Callable[[web.Request], Awaitable[web.StreamResponse]]",
                cm_service,
            ),
        ),
        web.get("/services/cm-service/cleanup", cm_service_cleanup),
        web.get(
            "/services/request-based-websocket",
            cast(
                "Callable[[web.Request], Awaitable[web.StreamResponse]]",
                websocket_request_based_service,
            ),
        ),
    ]
)


def main() -> None:
    container = Container()
    add_request_context(container)
    container.add_context_manager(CMService, scope=Scope.REQUEST)

    settings = container.resolve(AioHttpE2ESettings)
    web.run_app(app, host=settings.host, port=settings.port)


if __name__ == "__main__":
    main()



"""Non-component Annotated metadata is ignored for registration key normalization."""

from __future__ import annotations

from typing import Annotated, TypeAlias

from diwire import Component, Container

Priority: TypeAlias = Annotated[int, Component("priority")]
PriorityWithMeta: TypeAlias = Annotated[int, Component("priority"), "meta"]


def main() -> None:
    container = Container()
    container.add_instance(7, provides=int)
    container.add_instance(42, provides=Priority)

    plain = container.resolve(Annotated[int, "request-id"])
    component = container.resolve(PriorityWithMeta)

    print(f"plain={plain}")  # => plain=7
    print(f"component={component}")  # => component=42


if __name__ == "__main__":
    main()



"""Focused example: providers can read values from ``contextvars.ContextVar``."""

from __future__ import annotations

from contextvars import ContextVar

from diwire import Container, Lifetime, Scope

request_value_var: ContextVar[int] = ContextVar("request_value", default=0)


class RequestValue:
    def __init__(self, value: int) -> None:
        self.value = value


def build_request_value() -> RequestValue:
    return RequestValue(value=request_value_var.get())


def main() -> None:
    container = Container()
    container.add_factory(
        build_request_value,
        provides=RequestValue,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope(Scope.REQUEST) as request_scope:
        token = request_value_var.set(7)
        try:
            resolved = request_scope.resolve(RequestValue)
        finally:
            request_value_var.reset(token)

    print(f"provider_contextvar={resolved.value}")  # => provider_contextvar=7


if __name__ == "__main__":
    main()



from __future__ import annotations

import secrets

import django
from django.apps import apps
from django.conf import settings


def ensure_django_setup() -> None:
    if not settings.configured:
        settings.configure(
            ALLOWED_HOSTS=["*"],
            INSTALLED_APPS=[],
            MIDDLEWARE=[],
            SECRET_KEY=secrets.token_hex(16),
        )

    if not apps.ready:
        django.setup()
